/*..........................................................................*/
/*                                                                          */
/*      L a s t W a v e   K e r n e l   2 . 0 . 1                           */
/*                                                                          */
/*      Copyright (C) 1998-2003 Emmanuel Bacry.                             */
/*      email : lastwave@cmap.polytechnique.fr                              */
/*                                                                          */
/*..........................................................................*/
/*                                                                          */
/*      This program is a free software, you can redistribute it and/or     */
/*      modify it under the terms of the GNU General Public License as      */
/*      published by the Free Software Foundation; either version 2 of the  */
/*      License, or (at your option) any later version                      */
/*                                                                          */
/*      This program is distributed in the hope that it will be useful,     */
/*      but WITHOUT ANY WARRANTY; without even the implied warranty of      */
/*      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the       */
/*      GNU General Public License for more details.                        */
/*                                                                          */
/*      You should have received a copy of the GNU General Public License   */
/*      along with this program (in a file named COPYRIGHT);                */
/*      if not, write to the Free Software Foundation, Inc.,                */
/*      59 Temple Place, Suite 330, Boston, MA  02111-1307  USA             */
/*                                                                          */
/*..........................................................................*/




/*********************************************************
 *
 * LISTV contents
 *
 *********************************************************/


#include "lastwave.h"
#include <stdarg.h>
#include "signals.h"
#include "images.h"
#include "int_fsilist.h"



/* Type of a listv */
char *listvType = "&listv";


/*
 * Answers to the different print messages
 */

#define LVPrintLength 30

void PrintListv(LISTV lv)
{
  int n;
  
  Printf("%s {\n",listvType);
  for (n=0;n<lv->length;n++) {
    Printf("% 4d :  ",n);
    if (lv->values[n] == NULL) {
      Printf(LWDoubleFormat,lv->f[n]);
      Printf("\n");
    }
    else {
      Printf("%s\n",ToStrValue(lv->values[n],NO));
    }
  }
  Printf("}\n");
}

char *ToStrListv(LISTV lv,char flagShort)
{
  static char strShort[30];
  char *str,*str1,*str2;
  int n;
  int nAlloc;
  
  if (flagShort) {
    sprintf(strShort,"{length=%d}",lv->length);
    return(strShort);
  }
  
  nAlloc = 300;
  str = CharAlloc(nAlloc);
  TempPtr(str);
  if (LVPrintLength < lv->length) sprintf(str,"{length=%d; ",lv->length);
  else sprintf(str,"{");
  
  for (n=0;n<MIN(lv->length,LVPrintLength);n++) {
    if (lv->values[n] == NULL) {
      sprintf(strShort,LWDoubleFormat,lv->f[n]);
      str1 = strShort;
    }
    else str1 = ToStrValue(lv->values[n],YES);
    if (strlen(str)+strlen(str1) > nAlloc-20) {
      nAlloc += MAX(300,strlen(str1)+20);
      str2 = CharAlloc(nAlloc);
      TempPtr(str2);
      strcpy(str2,str);
      str = str2;
    }
    strcat(str,str1);    
    if (n != MIN(lv->length,LVPrintLength)-1) strcat(str," ");
  }
  if (LVPrintLength < lv->length) strcat(str,"...");
  strcat(str,"}");
  
  return(str);
}

void PrintInfoListv(LISTV lv)
{
  Printf("   length =  %d   [nAlloc=%d]\n",lv->length,lv->nAlloc);
}

/* 
 * Subroutine of SetListvField (just below)
 */
char *SetListv_(LISTV *plv,void **arg)
{
  extern void *SetFieldArg(VALUE value, void **arg);
  char *field = ARG_S_GetField(arg);
  FSIList *fsiList = ARG_S_GetFsiList(arg);
  char *type = ARG_S_GetRightType(arg);      
  LWFLOAT flt = ARG_S_GetRightFloat(arg);   
  VALUE val = ARG_S_GetRightValue(arg);
  char *equal = ARG_S_GetEqual(arg);
  VALUE *pValueRes = ARG_S_GetResPValue(arg);
  FSI_DECL;
  LISTV lv = *plv;
  LISTV lv1;
  int j,k,_iold;
  
  /*************************
   *
   * Case lv = lv1
   *
   *************************/
  if (fsiList == NULL && *equal == '=') {
    if (type != listvType) {
      SetErrorf("Right handside should be a listv");
      return(NULL);
    }
    lv1 = (LISTV) val;
    CopyListv(lv1,*plv);
    return(listvType);
  }
  
  /*************************
   *
   * Case lv[range] *= val
   *
   *************************/
  
  if (*equal != '=' && *equal != ':' && fsiList != NULL) {
    ARG_S_SetField(arg,NULL);
    ARG_S_SetFsiList(arg,NULL);
    FSI_FOR_START(fsiList);
    if (lv->values[_i] == NULL) {
      if (SetFloatField(&(lv->f[_i]),arg,0)==NULL) return(NULL);
    }
    else {
      if (SetFieldArg(lv->values[_i],arg) == NULL) return(NULL);      
    }
    FSI_FOR_END;
    return(listvType);
  }
  
  /*************************
   *
   * Case lv += val
   *
   *************************/
  if (*equal != '=' && *equal != ':') {
    
    switch(*equal) {
        
      case '+':
        if (val == NULL) AppendFloat2Listv(lv,flt);
        else if (GetTypeValue(val) != listvType) AppendValue2Listv(lv,val);
        else AppendListv2Listv(lv,CastValue(val,LISTV));
        return(listvType);
        
        case '*':
        if (val != NULL) {
          if (GetTypeValue(val) != numType) {
            SetErrorf("The right handside must be an integer");
            return(NULL);
          }
          flt = CastValue(val,NUMVALUE)->f;
        }
        if ((int) flt != flt) { 
          SetErrorf("The right handside must be an integer and not a LWFLOAT");
          return(NULL);
        }
        if (flt<0) { 
          SetErrorf("The right handside must be a positive integer");
          return(NULL);
        }
        lv1 = NewListv();
        CopyListv(lv,lv1);
        SetLengthListv(lv,0);
        MultListv(lv1,(int) flt, lv);
        DeleteListv(lv1);
        return(listvType);
        
        default : 
        SetErrorf("Inadequate syntax '%s'",equal);
        return(NULL);  
    }
    
  }
  
  /*************************
   *
   * Case lv[range]= val
   *
   *************************/
  if (*equal == '=') {
    
    /* case val is not a listv */
    if (type != listvType) {
      if (fsiList->nx != 1) {
        SetErrorf("Size of both handsides should match (left size = %d, right size =1)",fsiList->nx);
        return(NULL);
      }
      if (val == NULL) SetListvNthFloat(lv,(int) FSIArray((&(fsiList->fsi[0])),0),flt);
      else SetListvNthValue(lv,(int) FSIArray((&(fsiList->fsi[0])),0),val);
      return(listvType);
    }
    
    /* case val it is a listv */
    lv1 = CastValue(val,LISTV);
    if (fsiList->nx != lv1->length) {
      SetErrorf("Size of both handsides should match (left size = %d, right size = %d)",fsiList->nx,lv1->length);
      return(NULL);
    }
    FSI_FOR_START(fsiList);
    if (lv1->values[_k] != NULL) SetListvNthValue(lv,_i,lv1->values[_k]);
    else SetListvNthFloat(lv,_i,lv1->f[_k]);
    FSI_FOR_END;
    return(listvType);
  }  
  
  /*************************
   *
   * Case lv[range] := {val}
   *
   *************************/
  
  if (val == NULL || GetTypeValue(val)!=listvType) {
    SetErrorf("The syntax 'listv :=' should be followed by another 'listv'");
    return(NULL);
  }
  lv1 = CastValue(val,LISTV);
  if (lv1->length > 1) {
    SetErrorf("The syntax is 'listv := ' should be followed by another 'listv' of length 1 or 0");
    return(NULL);
  }   
  if (fsiList == NULL) {
    SetErrorf("Inadequate ':=' syntax");
    return(NULL);
  }
  
  if (lv1->length == 1) {
    if (lv1->values[0] == NULL) {
      FSI_FOR_START(fsiList);
      SetListvNthFloat(lv,_i,lv1->f[0]);
      FSI_FOR_END;
    }
    else {
      FSI_FOR_START(fsiList);
      SetListvNthValue(lv,_i,lv1->values[0]);
      FSI_FOR_END;
    }
    return(listvType);
  }
  
  /* We start by creating a new listv which is a copy of lv */
  lv1 = NewListv();
  CopyListv(lv,lv1);
  SetLengthListv(lv,lv1->length-fsiList->nx); 
  
  /* Then we peform the set */
  j = 0; /* Index in lv1 */
  k = 0; /* Index in lv */
  _iold = -1;
  FSI_FOR_START(fsiList);
  /* We check the fsiList is sorted */
  if (_i <= _iold) {
    SetErrorf("Sorry, this operation with non sorted indexes is not implemented");
    DeleteListv(lv1);
    return(NULL);
  }
  _iold = _i;
  memcpy(lv->values+k,lv1->values+j,(_i-j)*sizeof(VALUE));        
  memcpy(lv->f+k,lv1->f+j,(_i-j)*sizeof(LWFLOAT));        
  k+=_i-j;
  j=_i+1;
  FSI_FOR_END;
  memcpy(lv->values+k,lv1->values+j,(lv1->length-j)*sizeof(VALUE));
  memcpy(lv->f+k,lv1->f+j,(lv1->length-j)*sizeof(LWFLOAT));
  DeleteListv(lv1);
  return(listvType);
}



/*
 * Basic routine to deal with setting Listv fields
 */
void *SetListvField(LISTV *plv,void **arg)
{
  VALUE *pValueRes = ARG_S_GetResPValue(arg);
  char *field = ARG_S_GetField(arg);
  void *res;
  
  ARG_S_SetField(arg,NULL);
  res = SetListv_(plv,arg);
  ARG_S_SetField(arg,field);
  
  if (res == NULL) return(NULL);
  
  if (*pValueRes == NULL) *pValueRes = (VALUE) *plv;
  return(listvType);
}  



/* 
 * Routine to deal with setting of listv
 */

static char *doc = "{[*opt,...] [:]= (<value> | <listv>)} {Get/Set the listv elements}"; 

static void *SetExtractListvV(LISTV lv,void **arg)
{
  VALUE *pValueRes;
  
  /* doc */
  if (lv == NULL) return(doc);
  
  pValueRes = ARG_S_GetResPValue(arg);
  
  if (SetListv_(&lv,arg) == NULL) return(NULL);
  if (*pValueRes == NULL) *pValueRes = (VALUE) lv;
  return(listvType);
}


/*
 * The different extract options 
 */

static char *extractOptionsListv[] = {"*nolimit","*bperiodic","*bmirror","*bmirror1",NULL};
enum {
  FSIOptListvNoLimit = FSIOption1,
  FSIOptListvBPer = FSIOption2,
  FSIOptListvBMir = FSIOption3,
  FSIOptListvBMir1 = FSIOption4,
};

static char *doc1 = "{[*opt,...] [:]= (<value> | <listv>)} {Get/Set the element of a listv}"; 

/* 
 * The function that manages extraction of a listv
 */

static void *GetExtractListvV(VALUE val, void **arg)
{
  LISTV lv;
  FSIList *fsiList;
  LWFLOAT *pFlt;
  char **pStr;
  VALUE *pValue;
  
  int max;
  FSI_DECL;
  BorderType bt;
  char *type;
  LISTV lvResult;
  
  char flagListv;
  
  /* Doc */
  if (val == NULL) return(doc1);
  
  lv = (LISTV) val;
  fsiList = ARG_G_GetFsiList(arg);
  pFlt = ARG_G_GetResPFloat(arg);
  pStr = ARG_G_GetResPStr(arg);
  pValue = ARG_G_GetResPValue(arg);
  
  if (lv->length == 0) {
    if (fsiList != NULL && !(fsiList->options & FSIOptListvNoLimit) && fsiList->nx != 0) {    
      SetErrorf("No extraction on empty listv");
      Printf("NO\n");
      return(NULL);
    }
    *pValue = (VALUE) TNewListv();
    return(listvType);
  }

  /* case of an empty extraction */
  if (fsiList != NULL && fsiList->nx == 0) {
    *pValue = (VALUE) TNewListv();
    return(listvType);
  }
  
    
  if (fsiList->options & FSIOptListvBPer) bt = BorderPer;
  else if (fsiList->options & FSIOptListvBMir) bt = BorderMir;
  else if (fsiList->options & FSIOptListvBMir1) bt = BorderMir1;
  else bt = BorderNone;
  
  /* Get the max size */
  max = lv->length;
  
  /*
   * Case the result will be a simple val or a LWFLOAT
   */
  if (fsiList->options & (FSIOptListvBPer | FSIOptListvBMir | FSIOptListvBMir1) && fsiList->nx == 1 ||
      !(fsiList->options & (FSIOptListvBPer | FSIOptListvBMir | FSIOptListvBMir1)) && fsiList->nx1 == 1) {
    
    flagListv = (fsiList->fsi[0].type == FSISignal || fsiList->fsi[0].type == FSIRange || fsiList->size != 1);
    
    _i = (int) FSIArray((&(fsiList->fsi[0])),0);
    
    /* The loop in the case of *nolimit */
    if (fsiList->options & FSIOptListvNoLimit) {
      FSI_FOR_START(fsiList); 
      if (_i<0 || _i > max-1) continue;
      type=GetListvNth(lv,_i,pValue,pFlt);
      break;       
      FSI_FOR_END;
    }
    
    /* Other cases */
    else {
      switch (bt) {
        case BorderPer : type=GetListvNth(lv,BPER(_i,max),pValue,pFlt); break;
        case BorderMir : type=GetListvNth(lv,BMIR(_i,max),pValue,pFlt); break;
        case BorderMir1 : type=GetListvNth(lv,BMIR1(_i,max),pValue,pFlt); break;
        default : type=GetListvNth(lv,_i,pValue,pFlt);
      }
    }
    
    if (!flagListv) {
    if (*pValue) {
      AddRefValue(*pValue);
      TempValue(*pValue);
      return(type);
    }
    else return(numType);
  }
    else {
      LISTV lvResult = TNewListv();
      if (*pValue) AppendValue2Listv(lvResult,*pValue);
      else AppendFloat2Listv(lvResult,*pFlt);
      *pValue = (VALUE) lvResult;
      AddRefValue(*pValue);
      TempValue(*pValue);
      return(listvType);
    }
  }
  
  /*
   * Case the result will be a listv
   */
  
  /* Allocation of the result */
  lvResult = TNewListv();
  if (fsiList->options & (FSIOptListvBPer | FSIOptListvBMir | FSIOptListvBMir1)) SetLengthListv(lvResult,fsiList->nx);
  else SetLengthListv(lvResult,fsiList->nx1);
  lvResult->length = 0;
  
  /* The loop in the case of *nolimit */
  if (fsiList->options & FSIOptListvNoLimit) {
    FSI_FOR_START(fsiList); 
    if (_i<0 || _i > max-1) continue;
    GetListvNth(lv,_i,pValue,pFlt);
    FSI_FOR_END;
  }
  /* The loop in the generic case */
  else {
    FSI_FOR_START(fsiList); 
    switch (bt) {
      case BorderPer : GetListvNth(lv,BPER(_i,max),pValue,pFlt); break;
      case BorderMir : GetListvNth(lv,BMIR(_i,max),pValue,pFlt); break;
      case BorderMir1 : GetListvNth(lv,BMIR1(_i,max),pValue,pFlt); break;
      default : GetListvNth(lv,_i,pValue,pFlt);
        if (*pValue) AppendValue2Listv(lvResult,*pValue);
        else AppendFloat2Listv(lvResult,*pFlt);          
    }
    FSI_FOR_END;
  }
  
  *pValue = (VALUE) lvResult;
  
  return(listvType);
}

/*
 * Get the options for extraction (called for field NULL only)
 */

static char *optionDoc = "{{*nolimit,*bmirror,*bmirror1,*bperiodic} \
{*nolimit : indexes can be out of range} \
{*bperiodic : periodic border effect)} \
{*bmirror : mirror+periodic border effect (first and last elements are repeated)} \
{*bmirror1 : mirror+periodic border effect (first and last elements are NOT repeated)}\
}";

static void *GetExtractOptionsListvV(VALUE val, void **arg)
{
  /* doc */
  if (val == NULL) return(optionDoc);
  
  return(extractOptionsListv);
}


/*
 * Function to get the ExtractInfo for fields NULL
 */

static void *GetExtractInfoListvV(VALUE val, void **arg)
{
  static ExtractInfo extractInfo;
  static char flagInit = YES;
  
  
  LISTV lv = (LISTV) val;
  char *field = ARG_EI_GetField(arg);
  unsigned long *options = ARG_EI_GetPOptions(arg);
  
  /* If *bperiodic,... then *nolimit must be off */
  if (*options & (FSIOptListvBPer | FSIOptListvBMir | FSIOptListvBMir1)) *options &= ~FSIOptListvNoLimit;
  
  /* Init of the extraction info */
  if (flagInit) {
    extractInfo.xmin = 0;
    extractInfo.dx = 1;
    extractInfo.nSignals = 1;
    flagInit = NO;
  }
  
  /* Get the maximum index */
  if (lv->length == 0) {
    extractInfo.xmin = -1;
    extractInfo.xmax = 0;
    extractInfo.dx = 1;
    extractInfo.nSignals = 1;
    extractInfo.flags = EIIntIndex;
    return(&extractInfo);
  }
  else extractInfo.xmax = lv->length-1;
  
  /* '*nolimit' option : set some flags */
  if (*options & (FSIOptListvBPer | FSIOptListvBMir | FSIOptListvBMir1 | FSIOptListvNoLimit)) extractInfo.flags = EIIntIndex;
  else extractInfo.flags = EIErrorBound | EIIntIndex;
  
  return(&extractInfo);
}

/*
 * 'length/size' field
 */

static char *lengthDoc = "{[= <length>]} {Sets/Gets the length/size of a listv}";

static void * GetLengthListvV(VALUE val, void **arg)
{
  /* Documentation */
  if (val == NULL) return(lengthDoc);
  
  return(GetIntField(((LISTV) val)->length,arg));
}

/* Set the length of a listv (perform allocation) */
#define LISTVINCRLENGTH 2   /* If n element are required, we allocate n*LISTVINCRLENGTH */
#define LISTVMINLENGTH 10    /* The minimum number of elements of a listv */
void SetLengthListv(LISTV lv, int length)
{
  int i,l;
  
  if (length <= lv->nAlloc) {
    for (i=lv->length; i<length;i++) {
      if (lv->values[i] != NULL) {
        DeleteValue(lv->values[i]);
        lv->values[i] = NULL;
      }
      lv->f[i] = 0;
    }
  }
  else if (lv->nAlloc != 0) {
    l = (int) (length*LISTVINCRLENGTH);
    lv->f = Realloc(lv->f,sizeof(LWFLOAT)*l);
    lv->values = Realloc(lv->values,sizeof(VALUE)*l);
    for (i=lv->length; i<l;i++) {
      lv->values[i] = NULL;
      lv->f[i] = 0;
    }
    lv->nAlloc = l;
  }
  else {
    if (length < LISTVMINLENGTH) l = LISTVMINLENGTH;
    else l = (int) (length*LISTVINCRLENGTH);
    lv->f = Calloc(l,sizeof(LWFLOAT));
    lv->values = Calloc(l,sizeof(VALUE));
    lv->nAlloc = l;
  } 
  
  lv->length = length;
}

static void * SetLengthListvV(VALUE val, void **arg)
{
  LISTV lv;
  int length;
  
  /* doc */
  if (val == NULL) return(lengthDoc);
  
  lv = (LISTV) val;
  length = lv->length;
  
  if (SetIntField(&length,arg,FieldPositive)==NULL) return(NULL);
  
  SetLengthListv(lv,length);
}

/*
 * 'nAlloc' field
 */

static char *nallocDoc = "{} {Gets the allocation size of a listv}";

static void * GetNAllocListvV(VALUE val, void **arg)
{
  /* Documentation */
  if (val == NULL) return(nallocDoc);
  
  return(GetIntField(((LISTV) val)->nAlloc,arg));
}


/*
 * (Des)Allocation of a Listv
 */


/* Allocate a new listv */
LISTV NewListv(void)
{
  LISTV lv;
  
#ifdef DEBUGALLOC
  DebugType = "Listv";
#endif
  
  lv = Malloc(sizeof(struct listvValue));
  InitValue(lv,&tsListv);
  
  lv->length = 0;
  lv->nAlloc = 0;
  lv->values = NULL;
  lv->f = NULL;
  
  return(lv);
}

LISTV TNewListv(void)
{
  LISTV lv;
  
  lv = NewListv();
  
  TempValue( lv);
  
  return(lv);
}


/* Delete a listv */
void DeleteListv(LISTV lv)
{
  int i;
  
  RemoveRefValue( lv);
  if (lv->nRef>0) return;
  
  if (lv->f != NULL) Free(lv->f);
  
  if (lv->values != NULL) {
    
    for (i=0;i<lv->length;i++) {
      if(lv->values[i] != NULL) DeleteValue(lv->values[i]);
    }
    
    Free(lv->values);
  }
  
#ifdef DEBUGALLOC
  DebugType = "Listv";
#endif
  
  Free(lv);
}

/* Get the nth element wich should be a number */
LWFLOAT GetListvNthFloat(LISTV lv,int n)
{
  if (n>=lv->length || n< 0) Errorf("GetListvNthFloat() : index out of range (%d)",n);
  
  if (lv->values[n] == NULL) return(lv->f[n]);
  
  if (GetTypeValue(lv->values[n]) != numType) Errorf("GetListvNthFloat() : Bad element of type '%s'",GetTypeValue(lv->values[n]));
  return(CastValue(lv->values[n],NUMVALUE)->f);
}

char *GetListvNth(LISTV lv,int n,VALUE *val,LWFLOAT *f)
{  
  char *type;
  
  if (n>=lv->length || n< 0) Errorf("GetListvNth() : index out of range (%d)",n);
  
  if (lv->values[n] == NULL) {
    *val = NULL;
    *f = lv->f[n];
    return(numType);
  }
  
  *val = lv->values[n];
  
  type = GetTypeValue(*val);
  
  if (type == numType) *f = ((NUMVALUE) (*val))->f;
  
  return(type);
}


char *GetListvNthStr(LISTV lv,int n)
{
  if (n>=lv->length || n< 0) Errorf("GetListvNthStr() : index out of range (%d)",n);
  
  if (lv->values[n] == NULL) Errorf("GetListvNthStr() : expect a string, not a number");
  if (GetTypeValue(lv->values[n]) != strType) Errorf("GetListvNthStr() : expect a string, not a '%s'",GetTypeValue(lv->values[n]));
  
  return(CastValue(lv->values[n],STRVALUE)->str);
}

void SetListvNthFloat(LISTV lv,int n,LWFLOAT f)
{
  if (n>=lv->length || n < 0) Errorf("GetListvNth() : index out of range (%d)",n);
  
  if (lv->values[n] != NULL) {
    DeleteValue(lv->values[n]);
    lv->values[n] = NULL;
  }
  
  lv->f[n] = f;
  return;
}

void SetListvNthValue(LISTV lv,int n,VALUE val)
{
  extern char TestRecursiveListv(LISTV lv, LISTV lv1);
  
  if (n>=lv->length || n < 0) Errorf("GetListvNth() : index out of range (%d)",n);
  
  if (GetTypeValue(val) == listvType && TestRecursiveListv(lv,CastValue(val,LISTV)))
    Errorf("Sorry, recursive list are not allowed");
  
  if (lv->values[n] != NULL) {
    DeleteValue(lv->values[n]);
    lv->values[n] = NULL;
  }
  
  lv->values[n] = val;
  AddRefValue(val);
  
  return;
}

int AreEqualListv(LISTV lv1,LISTV lv2)
{
  int i;
  
  if (lv1->length != lv2->length) return(NO);
  
  for (i=0;i<lv1->length;i++) {
    if (lv1->values[i] != lv2->values[i]) return(NO);
    if (lv1->values[i] == NULL) {
      if (lv1->f[i] != lv2->f[i]) return(NO);
    }
    else {
      if (strType == GetTypeValue((VALUE) lv1->values[i])) {
        if (strcmp(CastValue(lv1->values[i],STRVALUE)->str,CastValue(lv2->values[i],STRVALUE)->str)) return(NO);
      }
    }
  }
  
  return(YES);
}

/* Copy a listv */
LISTV CopyListv(LISTV in,LISTV out)
{
  int i;
  
  if (in == out) return(out);
  
  if (out == NULL) out = NewListv();
  
  SetLengthListv(out,in->length);
  
  for (i=0;i<in->length;i++) {
    if (out->values[i] != NULL) DeleteValue(out->values[i]);
    out->values[i] = in->values[i];
    if (in->values[i]) AddRefValue(in->values[i]);
    else out->f[i] = in->f[i];
  }
  
  return(out);
}

/* Clear a listv */
void ClearListv(LISTV lv)
{
  int i;
  
  if (lv->f != NULL) Free(lv->f);
  
  if (lv->values != NULL) {
    
    for (i=0;i<lv->length;i++) {
      if(lv->values[i] != NULL) DeleteValue(lv->values[i]);
    }
    
    Free(lv->values);
  }
  
  lv->length = 0;
  lv->nAlloc = 0;
  lv->values = NULL;
  lv->f = NULL;
}

/* Concat 2 listv's */
void ConcatListv(LISTV lv1,LISTV lv2,LISTV lv3)
{
  int i;
  int l1 = lv1->length;
  int l2 = lv2->length;
  int l3 = l1+l2;
  
  SetLengthListv(lv3,l3);
  
  /* We first copy l2 */
  for (i=l1;i<l3;i++) {
    if (lv2->values[i-l1]) AddRefValue(lv2->values[i-l1]);
    else lv3->f[i] = lv2->f[i-l1];
    if (lv3->values[i] != NULL) DeleteValue(lv3->values[i]);
    lv3->values[i] = lv2->values[i-l1];
  }
  
  /* Then if necessary we copy l1 */
  if (lv1 != lv3) {
    for (i=0;i<l1;i++) {
      if (lv1->values[i]) AddRefValue(lv1->values[i]);
      else lv3->f[i] = lv1->f[i];
      if (lv3->values[i] != NULL) DeleteValue(lv3->values[i]);
      lv3->values[i] = lv1->values[i];
    }
  } 
}

/* Multiply a listv with an integer */
void MultListv(LISTV lv1, int n, LISTV lv3)
{
  int i,j,k,j0;
  int l1 = lv1->length;
  int l3 = l1*n;
  
  /*  if (lv1 == lv3) Errorf("MultListv() : the two listv should be different"); */
  
  SetLengthListv(lv3,l3);
  
  if (n == 0) return;
  
  if (lv1 == lv3) j0 = 1;
  else j0 = 0;  
  for (j=j0,k=j0*l1;j<n;j++) {
    for (i=0;i<l1;i++,k++) {
      if (lv1->values[i]) AddRefValue(lv1->values[i]);
      else lv3->f[k] = lv1->f[i];
      if (lv3->values[k] != NULL) DeleteValue(lv3->values[k]);
      lv3->values[k] = lv1->values[i];
    }
  }
}

/* Test if lv1 can be added as an element of lv */
char TestRecursiveListv(LISTV lv, LISTV lv1)
{
  int i;
  
  if (lv == lv1) return(YES);
  
  for (i=0;i<lv1->length;i++) {
    if (lv1->values[i]==NULL || GetTypeValue(lv1->values[i]) != listvType) continue;
    if (TestRecursiveListv(lv,CastValue(lv1->values[i],LISTV))) return(YES);
  }
  return(NO);
}

/* Append a Variable Content at the end of a listv */
void AppendValue2Listv (LISTV lv, VALUE val)
{
  if (GetTypeValue(val) == listvType && TestRecursiveListv(lv,CastValue(val,LISTV)))
    Errorf("Sorry, recursive list are not allowed");
  
  SetLengthListv(lv,lv->length+1);
  val = ValueOf(val);
  lv->values[lv->length-1] = val;
  AddRefValue(val);
}

/* Append a LWFLOAT at the end of a listv */
void AppendFloat2Listv (LISTV lv, LWFLOAT f)
{
  SetLengthListv(lv,lv->length+1);
  lv->values[lv->length-1] = NULL;
  lv->f[lv->length-1] = f;
}

/* Append a LWFLOAT array at the end of a listv */
void AppendFloatArray2Listv (LISTV lv, LWFLOAT *f, int n)
{
  int i,j;
  
  if (n <= 0) return;
  
  j=lv->length;
  SetLengthListv(lv,lv->length+n);
  for (i=0;i<n;i++,j++) {
    lv->values[j] = NULL;
    lv->f[j] = f[i];
  }
}

/* Append an int the end of a listv */
void AppendInt2Listv (LISTV lv, int i)
{
  SetLengthListv(lv,lv->length+1);
  lv->values[lv->length-1] = NULL;
  lv->f[lv->length-1] = i;
}

/* Append a LWFLOAT array at the end of a listv */
void AppendIntArray2Listv (LISTV lv, int *i, int n)
{
  int i1,j1;
  
  j1=lv->length;
  SetLengthListv(lv,lv->length+n);
  for (i1=0;i1<n;i1++,j1++) {
    lv->values[j1] = NULL;
    lv->f[j1] = i[i1];
  }
}

/* Append a string to the end of a listv */
void AppendStr2Listv (LISTV lv, char *str)
{
  STRVALUE sc;
  
  sc = NewStrValue();
  SetStrValue(sc,str);
  SetLengthListv(lv,lv->length+1);
  lv->values[lv->length-1] = (VALUE) sc;
}

/* The result is supposed to contain a list and we just want to append a string */
void AppendStr2Listvf(LISTV lv, char *format,...)
{
  Errorf("AppendStr2Listvf pas fait");
}

/* Append a listv to the end of a listv */
void AppendListv2Listv (LISTV lv, LISTV lv1)
{
  int i,j;
  int l1=lv1->length;  
  
  j = lv->length;
  SetLengthListv(lv,lv->length+lv1->length);
  for (i=0;i<l1;i++) {
    if (lv1->values[i] == NULL) {
      lv->values[j] = NULL;
      lv->f[j] = lv1->f[i];
    }
    else {
      if (GetTypeValue(lv1->values[i]) == listvType && TestRecursiveListv(lv,CastValue(lv1->values[i],LISTV))) Errorf("Sorry, recursive list are not allowed");
      lv->values[j] = lv1->values[i];
      AddRefValue(lv1->values[i]);
    }
    j++;
  }
}

extern void ApplyProc2Listv(LWPROC proc, LISTV lv);

static char _theFlagInverse=0;

static int qsortcmpstr(const void *s1,const void *s2)
{
  switch(_theFlagInverse) {
    case 0:
      return(strcmp((*((STRVALUE *) s1))->str,(*((STRVALUE *) s2))->str));
    default :
      return(strcmp((*((STRVALUE *) s2))->str,(*((STRVALUE *) s1))->str));
  }
}

static int qsortcmpnum(const void *s1,const void *s2)
{
  switch(_theFlagInverse) {
    case 0: 
      return((*((LWFLOAT *) s1)) >= (*((LWFLOAT *) s2)) ? 1 : -1);
    default :
      return((*((LWFLOAT *) s2)) >= (*((LWFLOAT *) s1)) ? 1 : -1);
  }
}

static LWPROC _theProc=NULL;
static LISTV _theLv= NULL;

static int qsortcmpvc(const void *s1,const void *s2)
{
  _theLv->values[0] = (*((VALUE *) s1));
  _theLv->values[1] = (*((VALUE *) s2));
  
  ApplyProc2Listv(_theProc,_theLv);
  
  switch(_theFlagInverse) {
    case 0:
      return(GetResultInt());
    default :
      return(-GetResultInt());
  }
}

LISTV SortListv(LISTV lv, LWPROC proc,char flagInverse)
{
  VALUE *vcarray;
  LWFLOAT *farray;
  int i;
  LISTV lv1;
  
  if (lv->length == 0) return(lv);
  
  _theFlagInverse = flagInverse;
  
  lv1 = TNewListv();
  CopyListv(lv,lv1);
  lv = lv1;
  
  if (proc != NULL) {
    vcarray = TMalloc(sizeof(VALUE)*lv->length);
    for (i=0;i<lv->length;i++) {
      vcarray[i] = lv->values[i];
      if (lv->values[i] != NULL && GetTrueTypeValue(lv->values[i]) == numType) {
        lv->f[i] = ((NUMVALUE) (lv->values[i]))->f;
        DeleteValue(lv->values[i]);
        lv->values[i] = NULL;
      }
      if (lv->values[i] == NULL) {
        vcarray[i] = (VALUE) TNewNumValue();
        ((NUMVALUE) (vcarray[i]))->f = lv->f[i];
      }
      else {
        vcarray[i] = lv->values[i];
      }
    }      
    _theProc = proc;
    if (_theLv == NULL) {
      _theLv = NewListv();
      SetLengthListv(_theLv,2);
    }
    qsort(vcarray,lv->length,sizeof(VALUE),&qsortcmpvc);
    for (i=0;i<lv->length;i++) {
      if (GetTypeValue(vcarray[i]) == numType) {
        lv->f[i] = CastValue(vcarray[i],NUMVALUE)->f;
        lv->values[i] = NULL;
      }
      else lv->values[i] = vcarray[i];
    }
    return(lv);
  }  
  
  
  if (lv->values[0] == NULL) {
    farray = TMalloc(sizeof(LWFLOAT)*lv->length);
    for (i=0;i<lv->length;i++) {
      if (lv->values[i] != NULL) break;
      farray[i] = lv->f[i];
    }
    if (i<lv->length) Errorf("Don't know how to sort");
    qsort(farray,lv->length,sizeof(LWFLOAT),&qsortcmpnum);
    for (i=0;i<lv->length;i++) {
      lv->f[i] = farray[i];
    }
    return(lv);
  }
  
  if (GetTypeValue(lv->values[0]) == strType) {
    vcarray = TMalloc(sizeof(VALUE)*lv->length);
    for (i=0;i<lv->length;i++) {
      if (lv->values[i] == NULL) break;
      if (GetTypeValue(lv->values[i]) != strType) break;
      vcarray[i] = lv->values[i];
    }
    if (i<lv->length) Errorf("Don't know how to sort");
    qsort(vcarray,lv->length,sizeof(VALUE),&qsortcmpstr);
    for (i=0;i<lv->length;i++) {
      lv->values[i] = vcarray[i];
    }
    return(lv);
  }
  SetErrorf("SortListv() : Don't know how to sort this listv");
  return(NULL);
}

void PrintListvColumn(LISTV lv,int colSize)
{
  int i,l;
  char *str;
  static NUMVALUE nc = NULL;
  
  if (lv->length == 0) return;
  
  if (nc == NULL) nc = NewNumValue();
  
  /* Computing the size of the columns if 'colSize' is not specified */
  if (colSize <= 0) {
    l = 0;
    for (i=0;i<lv->length;i++) {
      if (lv->values[i] != NULL) {
        if (GetTypeValue(lv->values[i]) == strType) str = ToStrValue(lv->values[i],NO);
        else str = ToStrValue(lv->values[i],YES);
      }
      else {
        nc->f = lv->f[i];
        str = ToStrValue(nc,YES);
      }
      if (*str == '\'' && str[strlen(str)-1] == '\'') {
        str[strlen(str)-1] = '\0';
        str++;
      }
      l = MAX(l,strlen(str));
    }
    colSize = l+2;
  }
  
  /* Then just print !! */
  PrintStrColumn(NULL,colSize);
  
  for (i=0;i<lv->length;i++) {
    if (lv->values[i] != NULL) {
      if (GetTypeValue(lv->values[i]) == strType) str = ToStrValue(lv->values[i],NO);
      else str = ToStrValue(lv->values[i],YES);
    }
    else {
      nc->f = lv->f[i];
      str = ToStrValue(nc,YES);
    }
    if (*str == '\'' && str[strlen(str)-1] == '\'') {
      str[strlen(str)-1] = '\0';
      str++;
    }
    PrintStrColumn(str,0);
  }
  
  PrintStrColumn(NULL,-1);
}


/*
 * Read a Listv made of strings or numbers ONLY from a column file
 */
typedef enum {
  skipCol = 0,
  strCol,
  numCol,
  signalCol,
  bestCol,
  lineNumberCol,
  anythingLeftCol
} colType;

/*
 * Useful function to skip multiple (windows or mac) new lines (returns the number of skipped new lines)
 */
static int SkipMultipleNewLines(FILE *stream)
{
  int n = 0;
  
  char c = fgetc(stream);
  while (1) {
    switch(c) {
      case '\n':
        n++;
        c = fgetc(stream);
        continue;
      case '\r' : 
        n++;
        c = fgetc(stream);
        if (c == '\n') c = fgetc(stream);
        continue;
      default :
        ungetc(c,stream);
    }
    break;
  }
  return(n);
}


/*
 * Read a multicolumn file as a listv
 *
 *    lv           : the listv to be read
 *    s            : the input stream
 *    nCols        : the expected number of columns
 *    ct           : array of size nCols made of colType element (to describe each column)
 *    nSkipLines   : number of lines to skip at the begining
 *    sizeToRead   : number of lines to read
 *    commentChars : the lines starting by one of these characters must be skipped
 *    separators1  : a string made of characters that act as separators
 *    spaceChars   : NOT TO BE CHANGED (default value is " \t") this string must contain the space character and any other character assimilated to the space character.
 *                   These will be used for skipping spaces at the begining of each column
 *    flagFake     : Only a single column in the listv with the 'n' 
 */
void CReadListvStream(LISTV lv,STREAM s, int nCols, colType *ct, int nSkipLines, int firstIndex,int sizeToRead, char *commentChars, char *separators1, char *spaceChars, char flagFake)
{
  FILE *stream;
  LISTV lv1;
  int i,j,n;
  LWFLOAT f;
  char str[10000],c,c1,flagSkipLine,*str1;
  int line,lineRead;
  SIGNAL ss;
  char formatAll[1000];
  char formatSkip[1000];
  char formatSpace[1000];
  char separatorsAll[1000];
  char ostr[1000];
  char *separators;
  int trueNCols;
  char flagFakeError;
  SIGNAL fakeSignal;
  
  // Compute the "true" number of columns (without the 'n')
  trueNCols = nCols;
  for (i=0;i<nCols;i++) if (ct[i] == lineNumberCol) trueNCols--;
  
  
  //
  // Init separators and formats :
  //     separatorsAll : characters that indicates we change column (including new lines)
  //     formatAll     : the corresponding format
  //     separators    : characters that indicates we change column (excluding new lines)
  //     formatSkip    : the format for skipping a column
  //     formatSpace   : the format for skipping spaces (and any characters considered as a space)
  //     
  separatorsAll[0] = '\n';
  separatorsAll[1] = '\r';
  separators = separatorsAll+2;
  strcpy(separators,separators1);
  
  strcpy(formatAll,"%[^");
  separators1 = separatorsAll;
  while (*separators1 != '\0') {
    switch(*separators1) 
    {
      case '%' : strcat(formatAll,"%%"); break;
      default : strncat(formatAll,separators1,1); break;
    }
    separators1++;
  }
  strcat(formatAll,"]");
  
  strcpy(formatSkip,"%*[");
  separators1 = separators;
  while (*separators1 != '\0') {
    switch(*separators1) 
    {
      case '%' : strcat(formatSkip,"%%"); break;
      default : strncat(formatSkip,separators1,1); break;
    }
    separators1++;
  }
  strcat(formatSkip,"]");
  
  strcpy(formatSpace,"%*[");
  str1 = spaceChars;
  while (*str1 != '\0') {
    switch(*str1) 
    {
      case '%' : strcat(formatSpace,"%%"); break;
      default : strncat(formatSpace,str1,1); break;
    }
    str1++;
  }
  strcat(formatSpace,"]");
  
  /* 
   * Some basic tests on the stream 
   */
  if (s->mode != StreamRead) Errorf("ReadListvStream() : The stream should be an input stream and not an output stream"); 
  if (s->stream == NULL) Errorf("ReadListvStream() : You cannot read a listv from standard input"); 
  stream = s->stream;   
  
  // Init the line number and the numberof lines read
  line = 1;
  lineRead = 1;
  
  // Skip the first lines
  while (!feof(stream) && nSkipLines != 0) {
    fscanf(stream,"%*[^\n\r]");
    nSkipLines--;
    line += SkipMultipleNewLines(stream);
  }
  if (nSkipLines != 0) Errorf("ReadListvStream() : too few lines in file");
  
  
  //
  // Allocation of the listv and of its elements (either signal's or listv's)
  //
  for (i = 0; i< nCols;i++) {
    if (ct[i] == skipCol) continue;
    if (ct[i] != signalCol && ct[i] != lineNumberCol && !flagFake) {
      lv1 = NewListv();
      AppendValue2Listv(lv,(VALUE) lv1);    
      DeleteValue(lv1);
    }
    else {
      ss = NewSignal();
      AppendValue2Listv(lv,(VALUE) ss);    
      DeleteSignal(ss);
    }
  }
  if (flagFake) {
    fakeSignal = NewSignal();
    AppendValue2Listv(lv,(VALUE) fakeSignal);    
    DeleteSignal(fakeSignal);
  }
  
  flagFakeError = 0;
  
  //******************************************************
  //
  // The main loop
  //
  //******************************************************
  if (firstIndex < 0) firstIndex = 0;
  while (!feof(stream) && sizeToRead != 0) {  // We loop till end of stream or sizeToRead is 0
    
    // Skip the line until firstIndex
    while(!feof(stream) && firstIndex > 0) {
      fscanf(stream,"%*[^\n\r]");
      firstIndex--;
      line += SkipMultipleNewLines(stream);
    }
    if (feof(stream)) continue;
    
    // We are going to read 1 line so we must decrease sizeToRead
    sizeToRead--;
    
    // Init
    flagSkipLine = NO;
    
    if (flagFake) AppendFloat2Signal(fakeSignal,0);
    
    //************************
    //
    // Loop on the column :  j is the index in the resulting listv and i the column index
    //
    //************************
    for (j = 0, i = 0; i<nCols;i++) {
      
      // If end of streqm is reached go out of the loop
      if (feof(stream)) break;
      
      // Case we must skip a line
      if (flagSkipLine) {
        fscanf(stream,"%*[^\n\r]");
        line += SkipMultipleNewLines(stream);
        i = nCols-1;
        if (flagFake) fakeSignal->size--;
        continue;
      }
      
      //
      // Here we read from the stream the next column
      // There are 2 cases : 
      //      anythingLeftCol --> we read the whole line
      //      otherwise       --> we just read the column (using the formatAll)
      // The resulting string (without leading spaces) will be stored in str1
      //
      str[0] = '\0';
      if (ct[i] == anythingLeftCol || ct[i] == lineNumberCol && ct[i+1] == anythingLeftCol) fscanf(stream,"%[^\n\r]",str);
      else fscanf(stream,formatAll,str);
      str1 = str;
      while (str1[0] != '\0' && strchr(spaceChars,str1[0])) str1++;
      
      // If we are in the first column and if we get a commentChar ==> we must skip the line
      if (i==0 && str1[0] != '\0' && commentChars != NULL && strchr(commentChars,str1[0])) {
        flagSkipLine = YES; 
        continue;
      }

      //
      // Different types of columns to read
      //

      if (ct[i] == lineNumberCol) { // Post processing if we record the line number 
        ss = (SIGNAL) lv->values[j];
        AppendFloat2Signal(ss,line);
        i++;
        j++;
      }
      
      switch(ct[i]) {
          
          //***************************
          // CASE OF A 'x' CHARACTER
          //***************************
        case skipCol : 
                    
          j-=1;  // The the listv index must not be incremented as it will be later (so we decrement it)
          
          break;
          
          //*****************************
          // CASE OF A 's' CHARACTER OR A '*' CHARACTER
          //*****************************
        case strCol : case anythingLeftCol : 
                    
          // We must store the string
          if (!flagFake) {
          lv1 = (LISTV) lv->values[j];
          AppendStr2Listv(lv1,str);
          }
            
          
          break;
          
          //*****************************
          // CASE OF A 'f' CHARACTER
          //*****************************
          case numCol : 
                    
          // We must read a number from this string
#ifdef NUMDOUBLE
          n = sscanf(str1,"%lf%s",&f,ostr);
#else
          n = sscanf(str1,"%f%s",&f,ostr);
#endif
          
          // If we read 0 argument ==> error
          if (n == 0) {
            if (flagFake) flagFakeError = 1;
            else Errorf("ReadListvStream() : line %d : column %d is not a number",line,i+1);
          }
          
          // If we read 2 arguments ==> error
          else if (n == 2) {            
            if (flagFake) flagFakeError = 1;
            else Errorf("ReadListvStream() : line %d : column %d is a number followed by a string",line,i+1);
          }
          
          // Oterwise, we must store the number
          else {
            if (!flagFake) {
          lv1 = (LISTV) lv->values[j];
          AppendFloat2Listv(lv1,f);
            }
          }
          
          break;          
          
          //*****************************
          // CASE OF A 'F' CHARACTER
          //*****************************
          case signalCol : 
                    
#ifdef NUMDOUBLE
          n = sscanf(str1,"%lf%s",&f,ostr);
#else
          n = sscanf(str1,"%f%s",&f,ostr);
#endif
          
          // If we read 0 argument ==> error
          if (n == 0) {
            if (flagFake) flagFakeError = 1;
            else Errorf("ReadListvStream() : line %d : column %d is not a number",line,i+1);
          }
          
          // If we read 2 arguments ==> error
          else if (n == 2) {            
            if (flagFake) flagFakeError = 1;
            else  Errorf("ReadListvStream() : line %d : column %d is a number followed by a string",line,i+1);
          }
          
          // Oterwise, we must store the string
          else {
            if (!flagFake) {
          ss = (SIGNAL) lv->values[j];
          AppendFloat2Signal(ss,f);
            }
          }
          
          break;
          
          //*****************************
          // CASE OF A '?' CHARACTER
          //*****************************
          case bestCol : 
                    
          // We try to read a number from this string
#ifdef NUMDOUBLE
          n = sscanf(str1,"%lf%s",&f,ostr);
#else
          n = sscanf(str1,"%f%s",&f,ostr);
#endif
          
          // If we read 0 or 2, we do not have a number, we have a string
          lv1 = (LISTV) lv->values[j];
          if (!flagFake) {
          if (n == 0 || n==2) AppendStr2Listv(lv1,str);
          // Otherwise, we have a number
          else AppendFloat2Listv(lv1,f);
          }
          
          break;
      }
      
      again0 :
      
      //
      // Dealing with the fake mode (and not skipping column)!
      //
      if (flagFake && ct[i] != skipCol) {
        
        // If no error --> we must store the 1 value
        if (!flagFakeError) {
          ss = (SIGNAL) lv->values[j];
          AppendFloat2Signal(ss,1);
          fakeSignal->Y[fakeSignal->size-1]++;
      }
      
        // If error --> store 0 and finish the line
        else {
          for(;i<nCols;i++,j++) {
            if (ct[i] == skipCol) {j--; continue;}
            ss = (SIGNAL) lv->values[j];
            AppendFloat2Signal(ss,0);
          }
          fscanf(stream,"%[^\n\r]",str);                  
          i--;
          j--;
          flagFakeError = 0;
        }        
      }
      
      
      // Increment the listv index
      j++;
      
      again : 
      
      // We haven't read the separator from the stream, we must take it out
      c = fgetc(stream);

      // We take out any extra space
      while (1) {
        c1 = fgetc(stream);
        if (!strchr(spaceChars,c1)) {ungetc(c1,stream); break;}
      }

      // If we reached a new line --> we change the separator
      if (c != '\r' && c != '\n' && (c1 == '\n' || c1 == '\r')) c = fgetc(stream);
          
      // Case the separator is a regular separator (i.e., excluding new lines) ==> we must check we do not expect the line to be over
      if (strchr(separators,c) != NULL) {
        if (nCols-1 == i) {
          if (flagFake) {
            fscanf(stream,"%*[^\n\r]");
            fakeSignal->Y[fakeSignal->size-1] = -1;
            goto again;
          }
          else Errorf("ReadListvStream() : line %d : more than %d columns",line,trueNCols);
        }
      }
      
      // Case the separator is a \n or \r ==> we must check we expect the line to be over
      else {
        if (nCols-1 != i) {
          // Looking for a * at the end
          if (ct[i+1] == anythingLeftCol && nCols-1 == i+1) {
            if (!flagFake) {
              lv1 = (LISTV) lv->values[j];
              AppendValue2Listv(lv1,nullValue);
            }
            else {
              ss = (SIGNAL) lv->values[j];
              AppendFloat2Signal(ss,1);              
              fakeSignal->Y[fakeSignal->size-1]++;
            }
            i++;
            j++;
          }
      else {
            if (flagFake) {
              flagFakeError = 1;
              i++;
              goto again0;
            }
            else Errorf("ReadListvStream() : line %d : less than %d columns",line,trueNCols);
          }
        }
        lineRead++;
        ungetc (c,stream);
        line += SkipMultipleNewLines(stream);
      }

    }
  }
  
}


/*
 * Write a Listv made of listv, strings or numbers/signals/images into a file (each element is stored one after the other)
 */

void WriteListvStream(LISTV lv,STREAM s, char flagBinary)
{
  int i;
  VALUE val;
  LWFLOAT f;
  SIGNAL sig;
  IMAGE im;
  LISTV lv1;
  char *type;
  char *str;
  FILE *stream;
  
  /* 
   * Some basic tests on the stream 
   */
  if (s->mode != StreamWrite) Errorf("WriteListvStream() : The stream should be an output stream and not an input stream"); 
  stream = s->stream;   
  
  /*
   * Writing a header
   */
  fprintf(stream,"LastWave 2.1 Listv\n"); 
  fprintf(stream,"length = %d\n",lv->length);
  if (flagBinary == NO) fprintf(stream,"binary no\n");
  else {
    if (IsCPULittleEndian) fprintf(stream,"binary little %d\n",sizeof(LWFLOAT));
    else fprintf(stream,"binary big %d\n",sizeof(LWFLOAT));
  }
  
  /* If length 0 --> that's it! */ 
  if (lv->length == 0) return;
  
  /* Otherwise we loop on the elements */
  for (i = 0;i<lv->length;i++) {
    
    /* Get the type of the element */
    type = GetListvNth(lv,i,&val,&f);
    fprintf(stream,"\n#>> Start Element #%d\n",i);
    fprintf(stream,"type = %s\n",type);
    
    /* Case of a string */
    if (type == strType) {
      str = GetListvNthStr(lv,i);
      fprintf(stream,"size = %d\n",strlen(str));
      fprintf(stream,"|%s|\n",str);
    }
    
    /* Case of a number */
    else if (type == numType) {
      if (!flagBinary) {
        f = GetListvNthFloat(lv,i);
#ifdef NUMDOUBLE	
        fprintf(stream,"%.16g\n",f);
#else
        fprintf(stream,"%.8g\n",f);
#endif
      }
      else {
        fwrite(&f,sizeof(LWFLOAT),1,stream);      
        fprintf(stream,"\n");
      }
    }
    
    /* Case of a signal */
    else if (type == signalType || type == signaliType) {
      sig = (SIGNAL) val;
      WriteSigStream(sig,s,flagBinary,"",YES,"","");
      fprintf(stream,"\n");
    }
    
    /* Case of an image */
    else if (type == imageType || type == imageiType) {
      im = (IMAGE) val;
      WriteImageStream(im,s,NO,YES,flagBinary,"","",NO,0,0);
      fprintf(stream,"\n");
    }
    
    /* Case of an listv */
    else if (type == listvType) {
      lv1 = (LISTV) val;
      WriteListvStream(lv1,s,flagBinary);
    }
    
    /* Other types --> error */
    else {
      Errorf("WriteListvStream() : Invalid element type '%s' : %s",type);  
      return;
    }
    
    fprintf(stream,"##<< Stop Element #%d\n",i);
  }
}   

/*
 * Read a Listv made of listv, strings or numbers/signals/images from a file (each element is stored one after the other)
 */

static void SkipCharsStream(FILE *stream, char flagNewLine);

static void Skip1LineStream(FILE *stream)
{
  fscanf(stream,"%*[^\n\r]");
  SkipCharsStream(stream,NO);
}

static char *Read1LineStream(FILE *stream)
{
  static char str[10000];
  
  fscanf(stream,"%[^\n\r]",&str);
  SkipCharsStream(stream,NO);
  
  return(str);
}

static void SkipCharsStream(FILE *stream, char flagNewLine)
{
  char c;
  
  while (1) {
    c = (char) fgetc(stream);
    if (c == ' ') {
      flagNewLine = NO;
      continue;
    }
    if  (c == '\n' || c == '\r') {
      flagNewLine = YES;
      continue;
    }
    if (c == '#' && flagNewLine) {
      Skip1LineStream(stream);
      flagNewLine = NO;
      continue;
    }
    ungetc(c,stream);
    break;
  }
}

void ReadListvStream(LISTV lv,STREAM s)
{
  char flagBinary;
  int i,n,size;
  LWFLOAT g;
  float f;
  double d;
  SIGNAL sig;
  IMAGE im;
  LISTV lv1;
  char type[100],*str;
  FILE *stream;
  int endmode,floatSize;
  extern void PrintSignal(SIGNAL signal);
  
  /* 
   * Some basic tests on the stream 
   */
  if (s->mode != StreamRead) Errorf("ReadListvStream() : The stream should be an input stream and not an output stream"); 
  stream = s->stream;   
  
  /*
   * Reading the header
   */
  
  /* Skip the first line */ 
  Skip1LineStream(stream);
  
  if (fscanf(stream,"length = %d",&n) != 1) Errorf("ReadListvStream() : Expect header line starting by the word 'length'");
  SkipCharsStream(stream,NO);
  SetLengthListv(lv,0);
  
  /* If length 0 --> that's it! */ 
  if (n == 0) return;
  
  str = Read1LineStream(stream);
  if (strncmp("binary ",str,7)) Errorf("ReadListvStream() : Expect header line starting by the word 'binary'");
  str += 7;
  if (!strcmp(str,"no")) flagBinary = NO;
  else {
    flagBinary = YES;
    if (!strncmp("little ",str,7)) {
      endmode = BinaryLittleEndian;
      str+=7;
    }
    else if (!strncmp("big ",str,4)) {
      endmode = BinaryBigEndian;
      str+=4;
    }
    else Errorf("ReadListvStream() : Error in header 'binary' line");
    if (sscanf(str,"%d",&floatSize) != 1) Errorf("ReadListvStream() : Error in header 'binary' line");
  }
  
  /* Otherwise we loop on the elements */
  for (i = 0;i<n;i++) {
    
    SkipCharsStream(stream,YES);  
    
    if (fscanf(stream,"type = %s",&type) != 1) Errorf("ReadListvStream() : Expect 'type' specification for element %d",i);
    SkipCharsStream(stream,NO);  
    
    
    /* Case of a string */
    if (!strcmp(type,strType)) {
      if (fscanf(stream,"size = %d",&size) != 1) Errorf("ReadListvStream() : size error while reading element %d which should be a string",i);
      SkipCharsStream(stream,NO);
      fgetc(stream);
      str = CharAlloc(size+1);
      if (size != 0 && fread(str,sizeof(char),size,stream) != size)
        Errorf("ReadListvStream() : Error while trying to read the element %d which should be a string",i);
      str[size] = '\0';  
      fgetc(stream);
      TempPtr(str);
      AppendStr2Listv(lv,str);      
    }
    
    /* Case of a number */
    else if (!strcmp(type,numType)) {
      if (!flagBinary) {
#ifdef NUMDOUBLE	
        fscanf(stream,"%lf",&g);
#else
        fscanf(stream,"%f",&g);
#endif
        AppendFloat2Listv(lv,g);
      }
      else {
        if (sizeof(float) == floatSize) {
          fread(&f,sizeof(float),1,stream);      
          if ((endmode == BinaryLittleEndian && IsCPUBigEndian) || ((endmode == BinaryBigEndian && IsCPULittleEndian)))
            BigLittleValues(&f,1,floatSize);
          AppendFloat2Listv(lv,f);  
        }
        else if (sizeof(double) == floatSize) {
          fread(&d,sizeof(double),1,stream);      
          if ((endmode == BinaryLittleEndian && IsCPUBigEndian) || ((endmode == BinaryBigEndian && IsCPULittleEndian)))
            BigLittleValues(&d,1,floatSize);
          AppendFloat2Listv(lv,(LWFLOAT) d);   
        }
        else Errorf("ReadListvStream() : sorry, do not know how to read binary numbers coded on %d bytes",floatSize);
      }
    }
    
    /* Case of a signal */
    else if (!strcmp(type,signalType) || !strcmp(type,signaliType)) {
      sig = TNewSignal();
      ReadSigStream(sig,s,-1,-1,0,0);
      AppendValue2Listv(lv,(VALUE) sig);
    }
    
    /* Case of an image */
    else if (!strcmp(type,imageType) || !strcmp(type,imageiType)) {
      im = TNewImage();
      ReadImageStream(im,s,-1,-1,NO,NO,NO,0);
      AppendValue2Listv(lv,(VALUE) im);
    }
    
    /* Case of an listv */
    else if (!strcmp(type,listvType)) {
      lv1 = TNewListv();
      ReadListvStream(lv1,s);
      AppendValue2Listv(lv,(VALUE) lv1);
    }
    
    /* Other types --> error */
    else {
      Errorf("ReadListvStream() : Invalid %dth element type '%s' : %s",i,type);  
      return;
    }
  }
}   


/*
 * Write a Listv made of strings or numbers ONLY into a column file
 *
 *   - each column will be separated by the string 'sep'
 *   - the 'formatArray' array indicates how the formatting of each column should be same syntax as printf)
 *     It must be either a NULL array (default format) or an array of size == the number of column.
 *     If a format is NULL then default foramt is used. 
 */

void CWriteListvStream(LISTV lv,STREAM s, char *sep, char **formatArray)
{
  int i,j,n,length;
  VALUE val;
  LWFLOAT f;
  char *type;
  LISTV lv1;
  SIGNAL sig;
  char *str;
  FILE *stream;
  char **typeArray,*format;
  
  /* 
   * Some basic tests on the stream 
   */
  if (s->mode != StreamWrite) Errorf("CWriteListvStream() : The stream should be an output stream and not an input stream"); 
  stream = s->stream;   
  
  if (lv->length == 0) return;
  i = 0;
  
  /* Let's check the size */
  length= -1;
  for (j=0;j<lv->length;j++) {
    if ((type = GetListvNth(lv,j,&val,&f)) == listvType) {
      lv1 = (LISTV) val;
      n = lv1->length;
    }
    else if (type == signaliType || type == signalType) {
      sig = (SIGNAL) val;
      n = sig->size;
    }
    else Errorf("CWriteListvStream() : The listv must be made of listv or signals only");
    if (length == -1) length = n;
    else if (length != n) Errorf("CWriteListvStream() : The listv/signals must have the same size");
  }
  
  /* Let's check the formatArray and convert it to a typeArray */
  if (formatArray == NULL) typeArray = NULL;
  else {
    typeArray = (char **) Malloc(sizeof(char *)*lv->length);
    TempPtr(typeArray);
    for (j=0;j<lv->length;j++) {
      if (formatArray[j] == NULL || !strcmp(formatArray[j],"")) typeArray[j] = NULL;
      else {
        
        /* Get the format main character */
        format = formatArray[j];
        if (strlen(format)<2 && format[0] != '%') Errorf("CWriteListvStream() : Bad format '%s'",format);
        format++;
        while(*format != '\0' && *format != '%' && !isalpha(*format)) format++;
        if (*format == '\0' || *format == '%') Errorf("CWriteListvStream() : Bad format '%s'",formatArray[j]);
        
        /* Case of an int */
        if (*format == 'd' || *format == 'i' || *format == 'o' ||
            *format == 'x' || *format == 'X' || *format == 'u') typeArray[j] = intType;
        
        /* Case of an double */
        else if (*format == 'f' || *format == 'e' || *format == 'E' ||
                 *format == 'g' || *format == 'G') typeArray[j] = floatType;
        
        /* Case of a string */
        else if (*format == 's') typeArray[j] = strType;
        
        /* Error */
        else Errorf("CWriteListvStream() : Bad format '%s'",formatArray[j]);
      }
    }
  }
  
  /* The loop on the lines */      
  while (1) {
    
    /* The loop on the columns */
    for (j=0;j<lv->length;j++) {
      
      /* Print separator if needed */
      if (j != 0) fprintf(stream,"%s",sep);
      
      /* Case the column is a list */
      if ((type = GetListvNth(lv,j,&val,&f)) == listvType) {
        /* Get the list and get the jth element */
        lv1 = (LISTV) val;
        type = GetListvNth(lv1,i,&val,&f);
        if (type == numType) {}
        else if (type == strType) str = GetStrFromStrValue((STRVALUE) val);
        else Errorf("CWriteListvStream() : listv must be made of floats/strings only");
      }
      
      /* case the column is a signal */
      else if (type == signaliType || type == signalType) {
        /* Get the signal and the jth element */
        sig = (SIGNAL) val;
        type = numType;
        f = sig->Y[i];
      }
      
      /*
       * case we must print a number 
       */
      if (type == numType) {
        /* Case no format specified */        
        if (typeArray == NULL || typeArray[j] == NULL || typeArray[j] == strType) {
#ifdef NUMDOUBLE      
          format = "%.16g";
#else
          format = "%.8g";
#endif
          fprintf(stream,format,f);
        }
        /* case of float */
        else if (typeArray[j] == floatType) fprintf(stream,formatArray[j],f);
        /* case of int */
#ifdef NUMDOUBLE      
        else fprintf(stream,formatArray[j],(long int) f);
#else
        else fprintf(stream,formatArray[j],(int) f);
#endif        
      }
      
      
      /*
       * case we must print a string 
       */
      else {
        if (typeArray == NULL || typeArray[j] == NULL || typeArray[j] != strType) format = "%s";
        else format = formatArray[j];
        fprintf(stream,format,str);
      }
      
    }
    
    /* Print a new line */
    fprintf(stream,"\n");
    
    /* Change line */
    i++;
    
    /* Should we stop ? */
    if (length == i) break;
  }
}


void C_Listv(char **argv)
{
  char *action,*sep,**formatArray,*type,*commentChars,*leadingChars,spaceChars[1000];
  LWPROC proc;
  LISTV lv,lv1,lv2;
  LWFLOAT f;
  VALUE val;
  int j,n,i;
  char flagInverse,opt;
  char *filename;
  colType ct[300];
  int nCols;
  int first, sizeToRead,skip;
  STREAM stream;
  char flagBinary;
  char *separators;
  char flagFake;
  
  argv = ParseArgv(argv,tWORD,&action,-1);
  
  if (!strcmp(action,"map")) {
    argv = ParseArgv(argv,tLISTV,&lv,-1);
    argv = ParseArgv(argv,tPROC,&proc,0);
    lv1 = TNewListv();
    lv2 = TNewListv();
    SetLengthListv(lv2,1);
    for (j=0;j<lv->length;j++) {
      GetListvNth(lv,j,&val,&f);
      if (val) SetListvNthValue(lv2,0,val);
      else SetListvNthFloat(lv2,0,f);
      ApplyProc2Listv(proc,lv2);
      if (GetResultType() != NULL) {
        if (GetResultType() == numType) AppendFloat2Listv(lv1,GetResultFloat());
        else AppendValue2Listv(lv1,GetResultValue());
      }
    }
    SetResultValue(lv1);
    return;
  }
  
  /* sort action */
  if (!strcmp(action,"sort")) {
    argv = ParseArgv(argv,tLISTV,&lv,-1);
    argv = ParseArgv(argv,tPROC_,NULL,&proc,-1);
    flagInverse = NO;
    while(opt = ParseOption(&argv)) { 
      switch(opt) {
        case 'i': flagInverse = YES; break;
        default: ErrorOption(opt);
      }
    }    
    NoMoreArgs(argv);
    lv = SortListv(lv,proc,flagInverse);
    if (lv == NULL) Errorf1("");
    SetResultValue(lv);  
    return;  
  }
  
  /* niceprint action */
  if (!strcmp(action,"niceprint")) {
    argv = ParseArgv(argv,tLISTV,&lv,-1);
    argv = ParseArgv(argv,tINT_,0,&n,0);
    PrintListvColumn(lv,n); 
    return;
  }
  
  /* cwrite action */
  if (!strcmp(action,"cwrite")) {
    argv = ParseArgv(argv,tLISTV,&lv,-1);
    
    /* Get the stream or the filaneme */
    filename = NULL;
    argv = ParseArgv(argv,tSTREAM_,NULL,&stream,-1);
    if (stream == NULL) {
      argv = ParseArgv(argv,tSTR,&filename,-1);
      stream = OpenFileStream(filename,"w");
      if (stream == NULL) Errorf("Can't open file '%s'",filename);
    }
    
    /* options */
    sep = "\t";
    formatArray = NULL;
    while(opt = ParseOption(&argv)) { 
      switch(opt) {
        case 's': argv = ParseArgv(argv,tSTR,&sep,-1); break;
        case 'f': 
          argv = ParseArgv(argv,tLISTV,&lv2,-1); 
          formatArray = (char **) Malloc(sizeof(char*)*lv2->length);
          TempPtr(formatArray);
          for (i=0;i<lv2->length;i++) {
            type = GetListvNth(lv2,i,&val,&f);
            if (type != strType) Errorf("Bad format type '%s' (should be string)",type);
            formatArray[i] = GetStrFromStrValue((STRVALUE) val);
          }        
          if (i != lv->length) Errorf("You must specify a format for each column (there are %d)",lv->length); 
          break;
          default: ErrorOption(opt);
      }
    }    
    NoMoreArgs(argv);
    
    CWriteListvStream(lv,stream,sep,formatArray);
    if (filename != NULL) CloseStream(stream);
    
    return;
  }
  
  /* write action */
  if (!strcmp(action,"write")) {
    argv = ParseArgv(argv,tLISTV,&lv,-1);
    
    /* Get the stream or the filaneme */
    filename = NULL;
    argv = ParseArgv(argv,tSTREAM_,NULL,&stream,-1);
    if (stream == NULL) {
      argv = ParseArgv(argv,tSTR,&filename,-1);
      stream = OpenFileStream(filename,"w");
      if (stream == NULL) Errorf("Can't open file '%s'",filename);
    }
    
    /* options */
    flagBinary = YES;
    while(opt = ParseOption(&argv)) { 
      switch(opt) {
        case 'a': flagBinary = NO; break;
        default: ErrorOption(opt);
      }
    }    
    NoMoreArgs(argv);
    
    WriteListvStream(lv,stream,flagBinary);
    
    if (filename != NULL) CloseStream(stream);
    
    return;
  }
  
  /* cread action */
  if (!strcmp(action,"cread")) {
    
    /* Get the stream or the filname */
    filename = NULL;
    argv = ParseArgv(argv,tSTREAM_,NULL,&stream,-1);
    if (stream == NULL) {
      argv = ParseArgv(argv,tSTR,&filename,-1);
      stream = OpenFileStream(filename,"r");
      if (stream == NULL) Errorf("Can't open file '%s'",filename);
    }
    
    nCols = 0;
    while (*argv != NULL) {
      if (!strcmp(*argv,"x")) ct[nCols++] = skipCol;
      else if (!strcmp(*argv,"s")) ct[nCols++] = strCol;
      else if (!strcmp(*argv,"f")) ct[nCols++] = numCol;
      else if (!strcmp(*argv,"*")) ct[nCols++] = anythingLeftCol;
      else if (!strcmp(*argv,"n")) ct[nCols++] = lineNumberCol;
      else if (!strcmp(*argv,"F")) ct[nCols++] = signalCol;
      else if (!strcmp(*argv,"?")) ct[nCols++] = bestCol;
      else break;
      argv++;
    }
    if (ct[nCols-1] == lineNumberCol) Errorf("column of type 'n' cannot be the last one");
    
    if (nCols == 0) ErrorUsage();
    
    /* options */
    skip = 0;
    first = 0;
    sizeToRead = -1;
    commentChars = NULL;
    leadingChars = NULL;
    separators = " \t";
    flagFake = 0;
    while(opt = ParseOption(&argv)) { 
      switch(opt) {
        case 'F': flagFake = 1; break;
        case 'l': argv = ParseArgv(argv,tSTR,&leadingChars,-1); break;
        case 'c': argv = ParseArgv(argv,tSTR,&commentChars,-1); break;
        case 's': argv = ParseArgv(argv,tINT,&skip,-1); break;
        case 'f': argv = ParseArgv(argv,tINT,&first,-1); break;
        case 'n': argv = ParseArgv(argv,tINT,&sizeToRead,-1); break;
        case 'S': argv = ParseArgv(argv,tSTR,&separators,-1); break;
        default: ErrorOption(opt);
      }
    }    
    
    NoMoreArgs(argv);
    
    spaceChars[0] = ' ';
    spaceChars[1] = '\t';
    spaceChars[2] = '\0';    
    if (leadingChars != NULL) strcat(spaceChars,leadingChars);
    
    lv = TNewListv();
    
    CReadListvStream(lv,stream,nCols,ct,skip,first,sizeToRead,commentChars,separators,spaceChars,flagFake);
    
    if (filename != NULL) CloseStream(stream);
    
    SetResultValue(lv);
    
    return;
  }
  
  /* read action */
  if (!strcmp(action,"read")) {
    
    /* Get the stream or the filaneme */
    filename = NULL;
    argv = ParseArgv(argv,tSTREAM_,NULL,&stream,-1);
    if (stream == NULL) {
      argv = ParseArgv(argv,tSTR,&filename,-1);
      stream = OpenFileStream(filename,"r");
      if (stream == NULL) Errorf("Can't open file '%s'",filename);
    }
    NoMoreArgs(argv);
    
    lv = TNewListv();
    
    ReadListvStream(lv,stream);
    if (filename != NULL) CloseStream(stream);
    
    SetResultValue(lv);
    
    return;
  }
  
  Errorf("Bad action '%s'",action);
}


/*
 * 'tosignal' field
 */

static char *tosignalDoc = "{} {Gets a signal made from a listv of numbers only}";

static void * GetTosignalListvV(LISTV lv, void **arg)
{
  SIGNAL sig;
  int i;
  
  /* Documentation */
  if (lv == NULL) return(tosignalDoc);
  
  sig = TNewSignal();
  SizeSignal(sig,lv->length,YSIG);
  
  for (i=0;i<lv->length;i++) {
    if (lv->values[i] == NULL) sig->Y[i] = lv->f[i];
    else if (GetTypeValue(lv->values[i]) != numType) {
      SetErrorf("A listv must contain only numbers to be converted to a signal");
      return(NULL);
    }
    else sig->Y[i] = CastValue(lv->values[i],NUMVALUE)->f;
  }
  
  return(GetValueField(sig,arg));
}


/*
 * The field list
 */
struct field fieldsListv[] = {
  
  "", GetExtractListvV, SetExtractListvV, GetExtractOptionsListvV, GetExtractInfoListvV,
  "length", GetLengthListvV, SetLengthListvV, NULL, NULL,
  "size", GetLengthListvV, SetLengthListvV, NULL, NULL,
  "nAlloc", GetNAllocListvV, NULL, NULL, NULL,
  "tosignal", GetTosignalListvV, NULL, NULL, NULL,
  
  NULL, NULL, NULL, NULL, NULL
};



/*
 * The type structure for LISTV
 */

TypeStruct tsListv = {
  
  "{{{&listv} {This type corresponds to list of values. \n \
  - ==,!= : test equality of all the elements \n \
  - Operator * : listv*n, repetition of the listv n times \n \
  - Operator + : listv+listv, Appending 2 listv \n \
  - Operator + : listv+value, (where value is not a listv), adding the value at the end of the listv  \n \
  - Constructor : syntax {value1 ... valueN}.}}}",  /* Documentation */
  
  &listvType,       /* The basic (unique) type name */
  NULL,     /* The GetType function */                       
  
  DeleteListv,     /* The Delete function */
  NewListv,     /* The New function */
  
  CopyListv,       /* The copy function */
  ClearListv,       /* The clear function */
  
  ToStrListv,       /* String conversion */
  PrintListv,   /* The Print function : print the object when 'print' is called */
  PrintInfoListv,   /* The PrintInfo function : called by 'info' */
  
  NULL,              /* The NumExtract function : used to deal with syntax like 10a */
  
  fieldsListv,      /* The list of fields */
};

